# stdlib
from collections import defaultdict, namedtuple
import time
import urlparse

# 3rd party
import requests

# project
from collector.checks import AgentCheck
from config import _is_affirmative
from common.util import headers


class NodeNotFound(Exception):
    pass


ESInstanceConfig = namedtuple(
    'ESInstanceConfig', [
        'shard_level_metrics',
        'pshard_stats',
        'cluster_stats',
        'password',
        'service_check_tags',
        'tags',
        'timeout',
        'url',
        'username',
    ])


class ESCheck(AgentCheck):
    SERVICE_CHECK_CONNECT_NAME = 'elasticsearch.can_connect'
    SERVICE_CHECK_CLUSTER_STATUS = 'elasticsearch.cluster_health'
    SERVICE_CHECK_SHARD_STATE = 'elasticsearch.shard.state'

    DEFAULT_TIMEOUT = 5

    # Pre aggregated metrics concerning only primary shards
    # Those dicts are populated in the ESCheck constructor
    PRIMARY_SHARD_METRICS = {
        'elasticsearch.indices.count': (
            'gauge', 'indices', lambda indices: len(indices)
        ),
        }
    PRIMARY_SHARD_METRICS_POST_1_0 = {}

    # Same kind of dicts for SHARD_LEVEL_METRICS
    SHARD_LEVEL_METRICS = {}
    SHARD_LEVEL_METRICS_POST_1_0 = {}

    # Shard-specific metrics
    SHARD_LEVEL_METRICS_SUFFIX = {
        ".docs.count": ("gauge", "docs.count"),
        ".docs.deleted": ("gauge", "docs.deleted"),
        ".store.size": ("gauge", "store.size_in_bytes"),
        ".indexing.index.total": ("gauge", "indexing.index_total"),
        ".indexing.index.time": ("gauge", "indexing.index_time_in_millis", lambda v: float(v)/1000),
        ".indexing.index.current": ("gauge", "indexing.index_current"),
        ".indexing.delete.total": ("gauge", "indexing.delete_total"),
        ".indexing.delete.time": ("gauge", "indexing.delete_time_in_millis", lambda v: float(v)/1000),
        ".indexing.delete.current": ("gauge", "indexing.delete_current"),
        ".get.total": ("gauge", "get.total"),
        ".get.time": ("gauge", "get.time_in_millis", lambda v: float(v)/1000),
        ".get.current": ("gauge", "get.current"),
        ".get.exists.total": ("gauge", "get.exists_total"),
        ".get.exists.time": ("gauge", "get.exists_time_in_millis", lambda v: float(v)/1000),
        ".get.missing.total": ("gauge", "get.missing_total"),
        ".get.missing.time": ("gauge", "get.missing_time_in_millis", lambda v: float(v)/1000),
        ".search.query.total": ("gauge", "search.query_total"),
        ".search.query.time": ("gauge", "search.query_time_in_millis", lambda v: float(v)/1000),
        ".search.query.current": ("gauge", "search.query_current"),
        ".search.fetch.total": ("gauge", "search.fetch_total"),
        ".search.fetch.time": ("gauge", "search.fetch_time_in_millis", lambda v: float(v)/1000),
        ".search.fetch.current": ("gauge", "search.fetch_current")
    }

    SHARD_LEVEL_METRICS_POST_1_0_SUFFIX = {
        ".merges.current": ("gauge", "merges.current"),
        ".merges.current.docs": ("gauge", "merges.current_docs"),
        ".merges.current.size": ("gauge", "merges.current_size_in_bytes"),
        ".merges.total": ("gauge", "merges.total"),
        ".merges.total.time": ("gauge", "merges.total_time_in_millis", lambda v: float(v)/1000),
        ".merges.total.docs": ("gauge", "merges.total_docs"),
        ".merges.total.size": ("gauge", "merges.total_size_in_bytes"),
        ".refresh.total": ("gauge", "refresh.total"),
        ".refresh.total.time": ("gauge", "refresh.total_time_in_millis", lambda v: float(v)/1000),
        ".flush.total": ("gauge", "flush.total"),
        ".flush.total.time": ("gauge", "flush.total_time_in_millis", lambda v: float(v)/1000)
    }

    STATS_METRICS = {  # Metrics that are common to all Elasticsearch versions
        "elasticsearch.docs.count": ("gauge", "indices.docs.count"),
        "elasticsearch.docs.deleted": ("gauge", "indices.docs.deleted"),
        "elasticsearch.store.size": ("gauge", "indices.store.size_in_bytes"),
        "elasticsearch.indexing.index.total": ("gauge", "indices.indexing.index_total"),
        "elasticsearch.indexing.index.time": ("gauge", "indices.indexing.index_time_in_millis", lambda v: float(v)/1000),
        "elasticsearch.indexing.index.current": ("gauge", "indices.indexing.index_current"),
        "elasticsearch.indexing.delete.total": ("gauge", "indices.indexing.delete_total"),
        "elasticsearch.indexing.delete.time": ("gauge", "indices.indexing.delete_time_in_millis", lambda v: float(v)/1000),
        "elasticsearch.indexing.delete.current": ("gauge", "indices.indexing.delete_current"),
        "elasticsearch.get.total": ("gauge", "indices.get.total"),
        "elasticsearch.get.time": ("gauge", "indices.get.time_in_millis", lambda v: float(v)/1000),
        "elasticsearch.get.current": ("gauge", "indices.get.current"),
        "elasticsearch.get.exists.total": ("gauge", "indices.get.exists_total"),
        "elasticsearch.get.exists.time": ("gauge", "indices.get.exists_time_in_millis", lambda v: float(v)/1000),
        "elasticsearch.get.missing.total": ("gauge", "indices.get.missing_total"),
        "elasticsearch.get.missing.time": ("gauge", "indices.get.missing_time_in_millis", lambda v: float(v)/1000),
        "elasticsearch.search.query.total": ("gauge", "indices.search.query_total"),
        "elasticsearch.search.query.time": ("gauge", "indices.search.query_time_in_millis", lambda v: float(v)/1000),
        "elasticsearch.search.query.current": ("gauge", "indices.search.query_current"),
        "elasticsearch.search.fetch.total": ("gauge", "indices.search.fetch_total"),
        "elasticsearch.search.fetch.time": ("gauge", "indices.search.fetch_time_in_millis", lambda v: float(v)/1000),
        "elasticsearch.search.fetch.current": ("gauge", "indices.search.fetch_current"),
        "elasticsearch.merges.current": ("gauge", "indices.merges.current"),
        "elasticsearch.merges.current.docs": ("gauge", "indices.merges.current_docs"),
        "elasticsearch.merges.current.size": ("gauge", "indices.merges.current_size_in_bytes"),
        "elasticsearch.merges.total": ("gauge", "indices.merges.total"),
        "elasticsearch.merges.total.time": ("gauge", "indices.merges.total_time_in_millis", lambda v: float(v)/1000),
        "elasticsearch.merges.total.docs": ("gauge", "indices.merges.total_docs"),
        "elasticsearch.merges.total.size": ("gauge", "indices.merges.total_size_in_bytes"),
        "elasticsearch.refresh.total": ("gauge", "indices.refresh.total"),
        "elasticsearch.refresh.total.time": ("gauge", "indices.refresh.total_time_in_millis", lambda v: float(v)/1000),
        "elasticsearch.flush.total": ("gauge", "indices.flush.total"),
        "elasticsearch.flush.total.time": ("gauge", "indices.flush.total_time_in_millis", lambda v: float(v)/1000),
        "elasticsearch.process.open_fd": ("gauge", "process.open_file_descriptors"),
        "elasticsearch.transport.rx_count": ("gauge", "transport.rx_count"),
        "elasticsearch.transport.tx_count": ("gauge", "transport.tx_count"),
        "elasticsearch.transport.rx_size": ("gauge", "transport.rx_size_in_bytes"),
        "elasticsearch.transport.tx_size": ("gauge", "transport.tx_size_in_bytes"),
        "elasticsearch.transport.server_open": ("gauge", "transport.server_open"),
        "elasticsearch.thread_pool.bulk.active": ("gauge", "thread_pool.bulk.active"),
        "elasticsearch.thread_pool.bulk.threads": ("gauge", "thread_pool.bulk.threads"),
        "elasticsearch.thread_pool.bulk.queue": ("gauge", "thread_pool.bulk.queue"),
        "elasticsearch.thread_pool.flush.active": ("gauge", "thread_pool.flush.active"),
        "elasticsearch.thread_pool.flush.threads": ("gauge", "thread_pool.flush.threads"),
        'elasticsearch.thread_pool.flush.rejected': ('rate', 'thread_pool.flush.rejected'),
        "elasticsearch.thread_pool.flush.queue": ("gauge", "thread_pool.flush.queue"),
        "elasticsearch.thread_pool.generic.active": ("gauge", "thread_pool.generic.active"),
        "elasticsearch.thread_pool.generic.threads": ("gauge", "thread_pool.generic.threads"),
        "elasticsearch.thread_pool.generic.queue": ("gauge", "thread_pool.generic.queue"),
        'elasticsearch.thread_pool.generic.rejected': ('rate', 'thread_pool.generic.rejected'),
        'elasticsearch.thread_pool.generic.completed': ('gauge', 'thread_pool.generic.completed'),
        "elasticsearch.thread_pool.get.active": ("gauge", "thread_pool.get.active"),
        "elasticsearch.thread_pool.get.threads": ("gauge", "thread_pool.get.threads"),
        "elasticsearch.thread_pool.get.queue": ("gauge", "thread_pool.get.queue"),
        'elasticsearch.thread_pool.get.rejected': ('rate', 'thread_pool.get.rejected'),
        'elasticsearch.thread_pool.get.completed': ('gauge', 'thread_pool.get.completed'),
        "elasticsearch.thread_pool.index.active": ("gauge", "thread_pool.index.active"),
        "elasticsearch.thread_pool.index.threads": ("gauge", "thread_pool.index.threads"),
        "elasticsearch.thread_pool.index.queue": ("gauge", "thread_pool.index.queue"),
        'elasticsearch.thread_pool.index.rejected': ('rate', 'thread_pool.index.rejected'),
        'elasticsearch.thread_pool.index.completed': ('gauge', 'thread_pool.index.completed'),
        #"elasticsearch.thread_pool.management.active": ("gauge", "thread_pool.management.active"),
        #"elasticsearch.thread_pool.management.threads": ("gauge", "thread_pool.management.threads"),
        #"elasticsearch.thread_pool.management.queue": ("gauge", "thread_pool.management.queue"),
        "elasticsearch.thread_pool.merge.active": ("gauge", "thread_pool.merge.active"),
        "elasticsearch.thread_pool.merge.threads": ("gauge", "thread_pool.merge.threads"),
        "elasticsearch.thread_pool.merge.queue": ("gauge", "thread_pool.merge.queue"),
        "elasticsearch.thread_pool.percolate.active": ("gauge", "thread_pool.percolate.active"),
        "elasticsearch.thread_pool.percolate.threads": ("gauge", "thread_pool.percolate.threads"),
        "elasticsearch.thread_pool.percolate.queue": ("gauge", "thread_pool.percolate.queue"),
        "elasticsearch.thread_pool.refresh.active": ("gauge", "thread_pool.refresh.active"),
        "elasticsearch.thread_pool.refresh.threads": ("gauge", "thread_pool.refresh.threads"),
        "elasticsearch.thread_pool.refresh.queue": ("gauge", "thread_pool.refresh.queue"),
        'elasticsearch.thread_pool.refresh.rejected': ('rate', 'thread_pool.refresh.rejected'),
        'elasticsearch.thread_pool.refresh.completed': ('gauge', 'thread_pool.refresh.completed'),
        "elasticsearch.thread_pool.search.active": ("gauge", "thread_pool.search.active"),
        "elasticsearch.thread_pool.search.threads": ("gauge", "thread_pool.search.threads"),
        "elasticsearch.thread_pool.search.queue": ("gauge", "thread_pool.search.queue"),
        'elasticsearch.thread_pool.search.rejected': ('rate', 'thread_pool.search.rejected'),
        'elasticsearch.thread_pool.search.completed': ('gauge', 'thread_pool.search.completed'),
        "elasticsearch.thread_pool.snapshot.active": ("gauge", "thread_pool.snapshot.active"),
        "elasticsearch.thread_pool.snapshot.threads": ("gauge", "thread_pool.snapshot.threads"),
        "elasticsearch.thread_pool.snapshot.queue": ("gauge", "thread_pool.snapshot.queue"),
        'elasticsearch.thread_pool.snapshot.rejected': ('rate', 'thread_pool.snapshot.rejected'),
        'elasticsearch.thread_pool.snapshot.completed': ('gauge', 'thread_pool.snapshot.completed'),
        "elasticsearch.http.current_open": ("gauge", "http.current_open"),
        "elasticsearch.http.total_opened": ("gauge", "http.total_opened"),
        "jvm.mem.heap_committed": ("gauge", "jvm.mem.heap_committed_in_bytes"),
        "jvm.mem.heap_used": ("gauge", "jvm.mem.heap_used_in_bytes"),
        "jvm.mem.non_heap_committed": ("gauge", "jvm.mem.non_heap_committed_in_bytes"),
        "jvm.mem.non_heap_used": ("gauge", "jvm.mem.non_heap_used_in_bytes"),
        #'jvm.mem.pools.young.used': ('gauge', 'jvm.mem.pools.young.used_in_bytes'),
        #'jvm.mem.pools.young.max': ('gauge', 'jvm.mem.pools.young.max_in_bytes'),
        #'jvm.mem.pools.old.used': ('gauge', 'jvm.mem.pools.old.used_in_bytes'),
        #'jvm.mem.pools.old.max': ('gauge', 'jvm.mem.pools.old.max_in_bytes'),
        #'jvm.mem.pools.survivor.used': ('gauge', 'jvm.mem.pools.survivor.used_in_bytes'),
        #'jvm.mem.pools.survivor.max': ('gauge', 'jvm.mem.pools.survivor.max_in_bytes'),
        'jvm.mem.heap_in_use': ('gauge', 'jvm.mem.heap_used_percent'),
        'jvm.mem.heap_max': ('gauge', 'jvm.mem.heap_max_in_bytes'),
        "jvm.threads.count": ("gauge", "jvm.threads.count"),
        "jvm.threads.peak_count": ("gauge", "jvm.threads.peak_count"),
        'elasticsearch.fs.total.total_in_bytes': ('gauge', 'fs.total.total_in_bytes'),
        'elasticsearch.fs.total.free_in_bytes': ('gauge', 'fs.total.free_in_bytes'),
        'elasticsearch.fs.total.available_in_bytes': ('gauge', 'fs.total.available_in_bytes'),
    }

    JVM_METRICS_POST_0_90_10 = {
        "jvm.gc.collectors.young.count": ("gauge", "jvm.gc.collectors.young.collection_count"),
        "jvm.gc.collectors.young.collection_time": ("gauge", "jvm.gc.collectors.young.collection_time_in_millis", lambda v: float(v)/1000),
        "jvm.gc.collectors.old.count": ("gauge", "jvm.gc.collectors.old.collection_count"),
        "jvm.gc.collectors.old.collection_time": ("gauge", "jvm.gc.collectors.old.collection_time_in_millis", lambda v: float(v)/1000)
    }

    JVM_METRICS_PRE_0_90_10 = {
        "jvm.gc.concurrent_mark_sweep.count": ("gauge", "jvm.gc.collectors.ConcurrentMarkSweep.collection_count"),
        "jvm.gc.concurrent_mark_sweep.collection_time": ("gauge", "jvm.gc.collectors.ConcurrentMarkSweep.collection_time_in_millis", lambda v: float(v)/1000),
        "jvm.gc.par_new.count": ("gauge", "jvm.gc.collectors.ParNew.collection_count"),
        "jvm.gc.par_new.collection_time": ("gauge", "jvm.gc.collectors.ParNew.collection_time_in_millis", lambda v: float(v)/1000),
        "jvm.gc.collection_count": ("gauge", "jvm.gc.collection_count"),
        "jvm.gc.collection_time": ("gauge", "jvm.gc.collection_time_in_millis", lambda v: float(v)/1000),
    }

    ADDITIONAL_METRICS_POST_0_90_5 = {
        "elasticsearch.search.fetch.open_contexts": ("gauge", "indices.search.open_contexts"),
        "elasticsearch.cache.filter.evictions": ("gauge", "indices.filter_cache.evictions"),
        "elasticsearch.cache.filter.size": ("gauge", "indices.filter_cache.memory_size_in_bytes"),
        "elasticsearch.id_cache.size": ("gauge", "indices.id_cache.memory_size_in_bytes"),
        "elasticsearch.fielddata.size": ("gauge", "indices.fielddata.memory_size_in_bytes"),
        "elasticsearch.fielddata.evictions": ("gauge", "indices.fielddata.evictions"),
    }

    ADDITIONAL_METRICS_PRE_0_90_5 = {
        "elasticsearch.cache.field.evictions": ("gauge", "indices.cache.field_evictions"),
        "elasticsearch.cache.field.size": ("gauge", "indices.cache.field_size_in_bytes"),
        "elasticsearch.cache.filter.count": ("gauge", "indices.cache.filter_count"),
        "elasticsearch.cache.filter.evictions": ("gauge", "indices.cache.filter_evictions"),
        "elasticsearch.cache.filter.size": ("gauge", "indices.cache.filter_size_in_bytes"),
    }

    ADDITIONAL_METRICS_POST_0_90_5_PRE_2_0 = {
    'elasticsearch.cache.filter.evictions': ('gauge', 'indices.filter_cache.evictions'),
    'elasticsearch.cache.filter.size': ('gauge', 'indices.filter_cache.memory_size_in_bytes'),
    'elasticsearch.id_cache.size': ('gauge', 'indices.id_cache.memory_size_in_bytes'),
    }

    ADDITIONAL_METRICS_POST_1_0_0 = {
        'elasticsearch.indices.translog.size_in_bytes': ('gauge', 'indices.translog.size_in_bytes'),
        'elasticsearch.indices.translog.operations': ('gauge', 'indices.translog.operations'),
    }

    # Stats are only valid for v1.x
    ADDITIONAL_METRICS_1_x = {
        'elasticsearch.fs.total.disk_reads': ('rate', 'fs.total.disk_reads'),
        'elasticsearch.fs.total.disk_writes': ('rate', 'fs.total.disk_writes'),
        'elasticsearch.fs.total.disk_io_op': ('rate', 'fs.total.disk_io_op'),
        'elasticsearch.fs.total.disk_read_size_in_bytes': ('gauge', 'fs.total.disk_read_size_in_bytes'),
        'elasticsearch.fs.total.disk_write_size_in_bytes': ('gauge', 'fs.total.disk_write_size_in_bytes'),
        'elasticsearch.fs.total.disk_io_size_in_bytes': ('gauge', 'fs.total.disk_io_size_in_bytes'),
    }

    ADDITIONAL_METRICS_POST_1_3_0 = {
        'elasticsearch.indices.segments.index_writer_memory_in_bytes': ('gauge', 'indices.segments.index_writer_memory_in_bytes'),
        #'elasticsearch.indices.segments.version_map_memory_in_bytes': ('gauge', 'indices.segments.version_map_memory_in_bytes'),
    }

    ADDITIONAL_METRICS_POST_1_4_0 = {
        'elasticsearch.indices.indexing.throttle_time': ('rate', 'indices.indexing.throttle_time_in_millis'),
        'elasticsearch.indices.query_cache.memory_size_in_bytes': ('gauge', 'indices.query_cache.memory_size_in_bytes'),
        'elasticsearch.indices.query_cache.hit_count': ('rate', 'indices.query_cache.hit_count'),
        'elasticsearch.indices.query_cache.miss_count': ('rate', 'indices.query_cache.miss_count'),
        'elasticsearch.indices.query_cache.evictions': ('rate', 'indices.query_cache.evictions'),
        'elasticsearch.indices.segments.index_writer_max_memory_in_bytes': ('gauge', 'indices.segments.index_writer_max_memory_in_bytes'),
        #'elasticsearch.indices.segments.fixed_bit_set_memory_in_bytes': ('gauge', 'indices.segments.fixed_bit_set_memory_in_bytes'),
        'elasticsearch.breakers.fielddata.estimated_size_in_bytes': ('gauge', 'breakers.fielddata.estimated_size_in_bytes'),
        #'elasticsearch.breakers.fielddata.overhead': ('gauge', 'breakers.fielddata.overhead'),
        'elasticsearch.breakers.fielddata.tripped': ('rate', 'breakers.fielddata.tripped'),
        #'elasticsearch.breakers.parent.estimated_size_in_bytes': ('gauge', 'breakers.parent.estimated_size_in_bytes'),
        #'elasticsearch.breakers.parent.overhead': ('gauge', 'breakers.parent.overhead'),
        #'elasticsearch.breakers.parent.tripped': ('rate', 'breakers.parent.tripped'),
        'elasticsearch.breakers.request.estimated_size_in_bytes': ('gauge', 'breakers.request.estimated_size_in_bytes'),
        #'elasticsearch.breakers.request.overhead': ('gauge', 'breakers.request.overhead'),
        'elasticsearch.breakers.request.tripped': ('rate', 'breakers.request.tripped'),
        #'elasticsearch.thread_pool.listener.active': ('gauge', 'thread_pool.listener.active'),
        #'elasticsearch.thread_pool.listener.threads': ('gauge', 'thread_pool.listener.threads'),
        #'elasticsearch.thread_pool.listener.queue': ('gauge', 'thread_pool.listener.queue'),
        #'elasticsearch.thread_pool.listener.rejected': ('rate', 'thread_pool.listener.rejected'),
    }

    ADDITIONAL_METRICS_POST_1_5_0 = {
        'elasticsearch.indices.recovery.current_as_source': ('gauge', 'indices.recovery.current_as_source'),
        'elasticsearch.indices.recovery.current_as_target': ('gauge', 'indices.recovery.current_as_target'),
        'elasticsearch.indices.recovery.throttle_time': ('rate', 'indices.recovery.throttle_time_in_millis'),
    }

    ADDITIONAL_METRICS_POST_1_6_0 = {
        #'elasticsearch.thread_pool.fetch_shard_started.active': ('gauge', 'thread_pool.fetch_shard_started.active'),
        #'elasticsearch.thread_pool.fetch_shard_started.threads': ('gauge', 'thread_pool.fetch_shard_started.threads'),
        #'elasticsearch.thread_pool.fetch_shard_started.queue': ('gauge', 'thread_pool.fetch_shard_started.queue'),
        #'elasticsearch.thread_pool.fetch_shard_started.rejected': ('rate', 'thread_pool.fetch_shard_started.rejected'),
        #'elasticsearch.thread_pool.fetch_shard_store.active': ('gauge', 'thread_pool.fetch_shard_store.active'),
        #'elasticsearch.thread_pool.fetch_shard_store.threads': ('gauge', 'thread_pool.fetch_shard_store.threads'),
        #'elasticsearch.thread_pool.fetch_shard_store.queue': ('gauge', 'thread_pool.fetch_shard_store.queue'),
        #'elasticsearch.thread_pool.fetch_shard_store.rejected': ('rate', 'thread_pool.fetch_shard_store.rejected'),
    }

    ADDITIONAL_METRICS_PRE_2_0 = {
        'elasticsearch.thread_pool.merge.active': ('gauge', 'thread_pool.merge.active'),
        'elasticsearch.thread_pool.merge.threads': ('gauge', 'thread_pool.merge.threads'),
        'elasticsearch.thread_pool.merge.queue': ('gauge', 'thread_pool.merge.queue'),
        'elasticsearch.thread_pool.merge.rejected': ('rate', 'thread_pool.merge.rejected'),
    }

    ADDITIONAL_METRICS_POST_2_0 = {
        # Some of these may very well exist in previous ES versions, but not worth the time/effort
        # to find where they were introduced
        #'elasticsearch.indices.query_cache.cache_size': ('gauge', 'indices.query_cache.cache_size'),
        #'elasticsearch.indices.query_cache.cache_count': ('rate', 'indices.query_cache.cache_count'),
        #'elasticsearch.indices.query_cache.total_count': ('rate', 'indices.query_cache.total_count'),
        'elasticsearch.indices.segments.doc_values_memory_in_bytes': ('gauge', 'indices.segments.doc_values_memory_in_bytes'),
        #'elasticsearch.indices.segments.norms_memory_in_bytes': ('gauge', 'indices.segments.norms_memory_in_bytes'),
        'elasticsearch.indices.segments.stored_fields_memory_in_bytes': ('gauge', 'indices.segments.stored_fields_memory_in_bytes'),
        #'elasticsearch.indices.segments.term_vectors_memory_in_bytes': ('gauge', 'indices.segments.term_vectors_memory_in_bytes'),
        'elasticsearch.indices.segments.terms_memory_in_bytes': ('gauge', 'indices.segments.terms_memory_in_bytes'),
        'elasticsearch.indices.request_cache.memory_size_in_bytes': ('gauge', 'indices.request_cache.memory_size_in_bytes'),
        'elasticsearch.indices.request_cache.evictions': ('rate', 'indices.request_cache.evictions'),
        'elasticsearch.indices.request_cache.hit_count': ('rate', 'indices.request_cache.hit_count'),
        'elasticsearch.indices.request_cache.miss_count': ('rate', 'indices.request_cache.miss_count'),
    }

    ADDITIONAL_METRICS_POST_2_1 = {
        'elasticsearch.indices.indexing.index_failed': ('rate', 'indices.indexing.index_failed'),
        'elasticsearch.thread_pool.force_merge.active': ('gauge', 'thread_pool.force_merge.active'),
        'elasticsearch.thread_pool.force_merge.threads': ('gauge', 'thread_pool.force_merge.threads'),
        'elasticsearch.thread_pool.force_merge.queue': ('gauge', 'thread_pool.force_merge.queue'),
        'elasticsearch.thread_pool.force_merge.rejected': ('rate', 'thread_pool.force_merge.rejected'),
    }

    ADDITIONAL_METRICS_PRE_5_0_0 = {
        'elasticsearch.thread_pool.percolate.active': ('gauge', 'thread_pool.percolate.active'),
        'elasticsearch.thread_pool.percolate.threads': ('gauge', 'thread_pool.percolate.threads'),
        'elasticsearch.thread_pool.percolate.queue': ('gauge', 'thread_pool.percolate.queue'),
        'elasticsearch.thread_pool.percolate.rejected': ('rate', 'thread_pool.percolate.rejected'),
        'elasticsearch.thread_pool.suggest.active': ('gauge', 'thread_pool.suggest.active'),
        'elasticsearch.thread_pool.suggest.threads': ('gauge', 'thread_pool.suggest.threads'),
        'elasticsearch.thread_pool.suggest.queue': ('gauge', 'thread_pool.suggest.queue'),
        'elasticsearch.thread_pool.suggest.rejected': ('rate', 'thread_pool.suggest.rejected'),
    }

    ADDITIONAL_METRICS_5_x = {
        'elasticsearch.fs.total.disk_io_op': ('rate', 'fs.io_stats.total.operations'),
        'elasticsearch.fs.total.disk_reads': ('rate', 'fs.io_stats.total.read_operations'),
        'elasticsearch.fs.total.disk_writes': ('rate', 'fs.io_stats.total.write_operations'),
        'elasticsearch.fs.total.disk_read_size_in_bytes': ('gauge', 'fs.io_stats.total.read_kilobytes'),
        'elasticsearch.fs.total.disk_write_size_in_bytes': ('gauge', 'fs.io_stats.total.write_kilobytes'),
        'elasticsearch.breakers.inflight_requests.tripped': ('gauge', 'breakers.in_flight_requests.tripped'),
        #'elasticsearch.breakers.inflight_requests.overhead': ('gauge', 'breakers.in_flight_requests.overhead'),
        'elasticsearch.breakers.inflight_requests.estimated_size_in_bytes': ('gauge', 'breakers.in_flight_requests.estimated_size_in_bytes'),
        'elasticsearch.search.scroll.total': ('gauge', 'indices.search.scroll_total'),
        'elasticsearch.search.scroll.time': ('gauge', 'indices.search.scroll_time_in_millis'),
        'elasticsearch.search.scroll.current': ('gauge', 'indices.search.scroll_current'),
    }

    ADDITIONAL_METRICS_PRE_6_3 = {
        'elasticsearch.thread_pool.bulk.active': ('gauge', 'thread_pool.bulk.active'),
        'elasticsearch.thread_pool.bulk.threads': ('gauge', 'thread_pool.bulk.threads'),
        'elasticsearch.thread_pool.bulk.queue': ('gauge', 'thread_pool.bulk.queue'),
        'elasticsearch.thread_pool.bulk.rejected': ('rate', 'thread_pool.bulk.rejected'),
        'elasticsearch.thread_pool.bulk.completed': ('rate', 'thread_pool.bulk.completed'),
    }

    ADDITIONAL_METRICS_POST_6_3 = {
        'elasticsearch.thread_pool.write.active': ('gauge', 'thread_pool.write.active'),
        'elasticsearch.thread_pool.write.threads': ('gauge', 'thread_pool.write.threads'),
        'elasticsearch.thread_pool.write.queue': ('gauge', 'thread_pool.write.queue'),
        'elasticsearch.thread_pool.write.rejected': ('rate', 'thread_pool.write.rejected'),
        'elasticsearch.thread_pool.write.completed': ('rate', 'thread_pool.write.completed'),
    }

    CLUSTER_HEALTH_METRICS = {
        "elasticsearch.number_of_nodes": ("gauge", "number_of_nodes"),
        "elasticsearch.number_of_data_nodes": ("gauge", "number_of_data_nodes"),
        "elasticsearch.active_primary_shards": ("gauge", "active_primary_shards"),
        "elasticsearch.active_shards": ("gauge", "active_shards"),
        "elasticsearch.relocating_shards": ("gauge", "relocating_shards"),
        "elasticsearch.initializing_shards": ("gauge", "initializing_shards"),
        "elasticsearch.unassigned_shards": ("gauge", "unassigned_shards"),
        'elasticsearch.delayed_unassigned_shards': ('gauge', 'delayed_unassigned_shards'),
        "elasticsearch.cluster_status": ("gauge", "status", lambda v: {"red": 0, "yellow": 1, "green": 2}.get(v, -1)),
    }

    CLUSTER_PENDING_TASKS = {
        "elasticsearch.pending_tasks_total": ("gauge", "pending_task_total"),
        "elasticsearch.pending_tasks_priority_high": ("gauge", "pending_tasks_priority_high"),
        "elasticsearch.pending_tasks_priority_urgent": ("gauge", "pending_tasks_priority_urgent"),
        'elasticsearch.pending_tasks_time_in_queue': ('gauge', 'pending_tasks_time_in_queue'),
    }

    SOURCE_TYPE_NAME = 'elasticsearch'

    def __init__(self, name, init_config, agentConfig, instances=None):

        # Let's construct the PRIMARY_SHARD_METRICS and SHARD_LEVEL_METRICS dicts
        for k, v in self.SHARD_LEVEL_METRICS_SUFFIX.iteritems():
            val = (v[0], '_all.primaries.{0}'.format(v[1]), v[2] if len(v) > 2 else None)
            self.PRIMARY_SHARD_METRICS['elasticsearch.primaries{0}'.format(k)] = val
            self.SHARD_LEVEL_METRICS['elasticsearch.shard{0}'.format(k)] = v

        for k, v in self.SHARD_LEVEL_METRICS_POST_1_0_SUFFIX.iteritems():
            val = (v[0], '_all.primaries.{0}'.format(v[1]), v[2] if len(v) > 2 else None)
            self.PRIMARY_SHARD_METRICS_POST_1_0['elasticsearch.primaries{0}'.format(k)] = val
            self.SHARD_LEVEL_METRICS_POST_1_0['elasticsearch.shard{0}'.format(k)] = v

        AgentCheck.__init__(self, name, init_config, agentConfig, instances)

        # Host status needs to persist across all checks
        self.cluster_status = {}

    def get_instance_config(self, instance):
        url = instance.get('url')
        if url is None:
            raise Exception("An url must be specified in the instance")

        pshard_stats = _is_affirmative(instance.get('pshard_stats', False))
        shard_level_metrics = _is_affirmative(instance.get('shard_level_metrics', False))

        cluster_stats = _is_affirmative(instance.get('cluster_stats', False))
        if 'is_external' in instance:
            cluster_stats = _is_affirmative(instance.get('is_external', False))

        # Support URLs that have a path in them from the config, for
        # backwards-compatibility.
        parsed = urlparse.urlparse(url)
        if parsed[2] != "":
            url = "%s://%s" % (parsed[0], parsed[1])
        port = parsed.port
        host = parsed.hostname
        service_check_tags = [
            'host:%s' % host,
            'port:%s' % port
        ]

        # Tag by URL so we can differentiate the metrics
        # from multiple instances
        tags = ['url:%s' % url]
        tags.extend(instance.get('tags', []))

        timeout = instance.get('timeout') or self.DEFAULT_TIMEOUT

        config = ESInstanceConfig(
            shard_level_metrics=shard_level_metrics,
            pshard_stats=pshard_stats,
            cluster_stats=cluster_stats,
            password=instance.get('password'),
            service_check_tags=service_check_tags,
            tags=tags,
            timeout=timeout,
            url=url,
            username=instance.get('username')
        )
        return config

    def check(self, instance):
        config = self.get_instance_config(instance)

        # Check ES version for this instance and define parameters
        # (URLs and metrics) accordingly
        version = self._get_es_version(config)

        health_url, stats_url, pshard_stats_url, pending_tasks_url, stats_metrics, \
            pshard_stats_metrics, shard_level_metrics = self._define_params(version,
                config.cluster_stats)

        # Load primary shards data
        if config.pshard_stats:
            pshard_stats_url = urlparse.urljoin(config.url, pshard_stats_url)
            pshard_stats_data = self._get_data(pshard_stats_url, config)
            self._process_pshard_stats_data(pshard_stats_data, config, pshard_stats_metrics)

        # Load shard-level metrics
        if config.shard_level_metrics:
            shard_level_url = urlparse.urljoin(config.url, pshard_stats_url) + '?level=shards'
            shard_level_data = self._get_data(shard_level_url, config)
            self._process_shard_level_data(shard_level_data, config, shard_level_metrics)

        # Load stats data.
        stats_url = urlparse.urljoin(config.url, stats_url)
        stats_data = self._get_data(stats_url, config)
        self._process_stats_data(stats_data, stats_metrics, config)

        # Load the health data.
        health_url = urlparse.urljoin(config.url, health_url)
        health_data = self._get_data(health_url, config)
        self._process_health_data(health_data, config)

        # Load the pending_tasks data.
        pending_tasks_url = urlparse.urljoin(config.url, pending_tasks_url)
        pending_tasks_data = self._get_data(pending_tasks_url, config)
        self._process_pending_tasks_data(pending_tasks_data, config)

        # If we're here we did not have any ES conn issues
        self.service_check(
            self.SERVICE_CHECK_CONNECT_NAME,
            AgentCheck.OK,
            tags=config.service_check_tags
        )

    def _get_es_version(self, config):
        """ Get the running version of elasticsearch.
        """
        try:
            data = self._get_data(config.url, config, send_sc=False)
            version = map(int, data['version']['number'].split('.')[0:3])
        except Exception, e:
            self.warning(
                "Error while trying to get Elasticsearch version "
                "from %s %s"
                % (config.url, str(e))
            )
            version = [1, 0, 0]

        self.service_metadata('version', version)
        self.log.debug("Elasticsearch version is %s" % version)
        return version

    def _define_params(self, version, cluster_stats):
        """ Define the set of URLs and METRICS to use depending on the
            running ES version.
        """
        pshard_stats_url = "/_stats"
        health_url = "/_cluster/health?pretty=true"

        if version >= [0, 90, 10]:
            # ES versions 0.90.10 and above
            pending_tasks_url = "/_cluster/pending_tasks?pretty=true"
            stats_url = "/_nodes/stats" if cluster_stats else "/_nodes/_local/stats"
            if version < [5, 0, 0]:
                # version 5 errors out if the `all` parameter is set
                stats_url += "?all=true"

            additional_metrics = self.JVM_METRICS_POST_0_90_10
        else:
            pending_tasks_url = None
            if cluster_stats:
                stats_url = "/_cluster/nodes/stats?all=true"
            else:
                stats_url = "/_cluster/nodes/_local/stats?all=true"

            additional_metrics = self.JVM_METRICS_PRE_0_90_10

        stats_metrics = dict(self.STATS_METRICS)
        stats_metrics.update(additional_metrics)

        if version >= [0, 90, 5]:
            # ES versions 0.90.5 and above
            stats_metrics.update(self.ADDITIONAL_METRICS_POST_0_90_5)
        else:
            # ES version 0.90.4 and below
            stats_metrics.update(self.ADDITIONAL_METRICS_PRE_0_90_5)

        # Version specific stats metrics about the primary shards
        pshard_stats_metrics = dict(self.PRIMARY_SHARD_METRICS)
        shard_level_metrics = dict(self.SHARD_LEVEL_METRICS)

        if version >= [1, 0, 0]:
            pshard_stats_metrics.update(self.PRIMARY_SHARD_METRICS_POST_1_0)
            shard_level_metrics.update(self.SHARD_LEVEL_METRICS_POST_1_0)
            stats_metrics.update(self.ADDITIONAL_METRICS_POST_1_0_0)

        if version < [2, 0, 0]:
            stats_metrics.update(self.ADDITIONAL_METRICS_PRE_2_0)
            if version >= [0, 90, 5]:
                stats_metrics.update(self.ADDITIONAL_METRICS_POST_0_90_5_PRE_2_0)
            if version >= [1, 0, 0]:
                stats_metrics.update(self.ADDITIONAL_METRICS_1_x)

        if version >= [1, 3, 0]:
            stats_metrics.update(self.ADDITIONAL_METRICS_POST_1_3_0)

        if version >= [1, 4, 0]:
            stats_metrics.update(self.ADDITIONAL_METRICS_POST_1_4_0)

        if version >= [1, 5, 0]:
            stats_metrics.update(self.ADDITIONAL_METRICS_POST_1_5_0)

        if version >= [1, 6, 0]:
            stats_metrics.update(self.ADDITIONAL_METRICS_POST_1_6_0)

        if version >= [2, 0, 0]:
            stats_metrics.update(self.ADDITIONAL_METRICS_POST_2_0)

        if version >= [2, 1, 0]:
            stats_metrics.update(self.ADDITIONAL_METRICS_POST_2_1)

        if version >= [5, 0, 0]:
            stats_metrics.update(self.ADDITIONAL_METRICS_5_x)

        if version < [5, 0, 0]:
            stats_metrics.update(self.ADDITIONAL_METRICS_PRE_5_0_0)

        if version >= [6, 3, 0]:
            stats_metrics.update(self.ADDITIONAL_METRICS_POST_6_3)
        else:
            stats_metrics.update(self.ADDITIONAL_METRICS_PRE_6_3)

        return health_url, stats_url, pshard_stats_url, pending_tasks_url, \
            stats_metrics, pshard_stats_metrics, shard_level_metrics

    def _get_data(self, url, config, send_sc=True):
        """ Hit a given URL and return the parsed json
        """
        # Load basic authentication configuration, if available.
        if config.username and config.password:
            auth = (config.username, config.password)
        else:
            auth = None

        try:
            resp = requests.get(
                url,
                timeout=config.timeout,
                headers=headers(self.agentConfig),
                auth=auth
            )
            resp.raise_for_status()
        except Exception as e:
            if send_sc:
                self.service_check(
                    self.SERVICE_CHECK_CONNECT_NAME,
                    AgentCheck.CRITICAL,
                    message="Error {0} when hitting {1}".format(e, url),
                    tags=config.service_check_tags
                )
            raise

        return resp.json()

    def _process_pending_tasks_data(self, data, config):
        p_tasks = defaultdict(int)

        for task in data.get('tasks', []):
            p_tasks[task.get('priority')] += 1

        node_data = {
            'pending_task_total':               sum(p_tasks.values()),
            'pending_tasks_priority_high':      p_tasks['high'],
            'pending_tasks_priority_urgent':    p_tasks['urgent'],
        }

        for metric in self.CLUSTER_PENDING_TASKS:
            # metric description
            desc = self.CLUSTER_PENDING_TASKS[metric]
            self._process_metric(node_data, metric, *desc, tags=config.tags)

    def _process_stats_data(self, data, stats_metrics, config):
        cluster_stats = config.cluster_stats
        for node_name in data['nodes']:
            node_data = data['nodes'][node_name]
            # On newer version of ES it's "host" not "hostname"
            node_hostname = node_data.get(
                'hostname', node_data.get('host', None))

            # Override the metric hostname if we're hitting an external cluster
            metric_hostname = node_hostname if cluster_stats else None

            for metric, desc in stats_metrics.iteritems():
                self._process_metric(
                    node_data, metric, *desc, tags=config.tags,
                    hostname=metric_hostname)

    def _process_pshard_stats_data(self, data, config, pshard_stats_metrics):
        for metric, desc in pshard_stats_metrics.iteritems():
            self._process_metric(data, metric, *desc, tags=config.tags)

    def _process_shard_level_data(self, data, config, shard_level_metrics):
        for index_name, index_data in data['indices'].iteritems():
            for i_str, pshard_and_replicas in index_data['shards'].iteritems():
                # Do we have more than one replica for this shard ?
                count_replicas = len(pshard_and_replicas) > 2
                replica_number = 0

                for shard in pshard_and_replicas:
                    # Let's get the host tag
                    node = shard['routing']['node']

                    # Let's compute our shard name
                    shard_role = "replica"
                    if shard['routing']['primary']:
                        shard_name = 'P' + i_str
                        shard_role = "primary"
                    elif count_replicas:
                        replica_number += 1
                        shard_name = 'R{0}_{1}'.format(i_str, replica_number)
                    else:
                        shard_name = 'R' + i_str

                    # Let's add some interesting tags that will enable us to
                    # slice and dice as we wish in Cloudinsight
                    es_shard_tags = ["es_node:{0}".format(node),
                                     "es_shard:{0}".format(shard_name),
                                     "es_index:{0}".format(index_name),
                                     "es_role:{0}".format(shard_role),
                                     "shard_specific"]

                    # Let's send a good old service check
                    if shard['routing']['state'] == 'STARTED':
                        state = AgentCheck.OK
                        sc_msg = "Shard is running"
                    elif shard['routing']['state'] == 'INITIALIZING':
                        state = AgentCheck.WARNING
                        sc_msg = ("Shard is currently being initialized. It should turn green "
                                  "(or red...) soon")
                    else:
                        state = AgentCheck.CRITICAL
                        sc_msg = "Shard is unassigned. Do you have enough nodes in your cluster ? "

                    self.service_check(
                        self.SERVICE_CHECK_SHARD_STATE,
                        state,
                        message=sc_msg,
                        tags=config.service_check_tags + es_shard_tags
                    )

                    # And let's finally send our metric !
                    for metric, desc in shard_level_metrics.iteritems():
                        self._process_metric(shard, metric, *desc, tags=config.tags +
                                             es_shard_tags)

    def _process_metric(self, data, metric, xtype, path, xform=None,
                        tags=None, hostname=None):
        """data: dictionary containing all the stats
        metric: agent metric
        path: corresponding path in data, flattened, e.g. thread_pool.bulk.queue
        xfom: a lambda to apply to the numerical value
        """
        value = data

        # Traverse the nested dictionaries
        for key in path.split('.'):
            if value is not None:
                value = value.get(key, None)
            else:
                break

        if value is not None:
            if xform:
                value = xform(value)
            if xtype == "gauge":
                self.gauge(metric, value, tags=tags, hostname=hostname)
            else:
                self.rate(metric, value, tags=tags, hostname=hostname)
        else:
            self._metric_not_found(metric, path)

    def _process_health_data(self, data, config):
        if self.cluster_status.get(config.url) is None:
            self.cluster_status[config.url] = data['status']
            if data['status'] in ["yellow", "red"]:
                event = self._create_event(data['status'], tags=config.tags)
                self.event(event)

        if data['status'] != self.cluster_status.get(config.url):
            self.cluster_status[config.url] = data['status']
            event = self._create_event(data['status'], tags=config.tags)
            self.event(event)

        for metric, desc in self.CLUSTER_HEALTH_METRICS.iteritems():
            self._process_metric(data, metric, *desc, tags=config.tags)

        # Process the service check
        cluster_status = data['status']
        if cluster_status == 'green':
            status = AgentCheck.OK
            data['tag'] = "OK"
        elif cluster_status == 'yellow':
            status = AgentCheck.WARNING
            data['tag'] = "WARN"
        else:
            status = AgentCheck.CRITICAL
            data['tag'] = "ALERT"

        msg = "{tag} on cluster \"{cluster_name}\" "\
              "| active_shards={active_shards} "\
              "| initializing_shards={initializing_shards} "\
              "| relocating_shards={relocating_shards} "\
              "| unassigned_shards={unassigned_shards} "\
              "| timed_out={timed_out}" \
              .format(**data)

        self.service_check(
            self.SERVICE_CHECK_CLUSTER_STATUS,
            status,
            message=msg,
            tags=config.service_check_tags
        )

    def _metric_not_found(self, metric, path):
        self.log.debug("Metric not found: %s -> %s", path, metric)

    def _create_event(self, status, tags=None):
        hostname = self.hostname.decode('utf-8')
        if status == "red":
            alert_type = "error"
            msg_title = "%s is %s" % (hostname, status)

        elif status == "yellow":
            alert_type = "warning"
            msg_title = "%s is %s" % (hostname, status)

        else:
            # then it should be green
            alert_type = "success"
            msg_title = "%s recovered as %s" % (hostname, status)

        msg = "ElasticSearch: %s just reported as %s" % (hostname, status)

        return {
            'timestamp': int(time.time()),
            'event_type': 'elasticsearch',
            'host': hostname,
            'msg_text': msg,
            'msg_title': msg_title,
            'alert_type': alert_type,
            'source_type_name': "elasticsearch",
            'event_object': hostname,
            'tags': tags
        }
