# stdlib
import os.path
import unittest

# 3rd party
import mock
from nose.plugins.attrib import attr

# project
from common.utils.flare import Flare


def get_mocked_config():
    return {
        'license_key': 'LICENSEKEY',
        'ci_url': 'http://cidc.aiops.com',
    }


def get_mocked_version():
    return '6.6.6'


def get_mocked_temp():
    return os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        'fixtures',
        'flare'
    )


def mocked_strftime(t):
    return '1'


class FakeResponse(object):
    def __init__(self, status_code=200):
        self.status_code = status_code
        self.text = '{"case_id":1337}'

    def json(self):
        return {'case_id': 1337}

    def raise_for_status(self):
        return None


class FlareTest(unittest.TestCase):

    @mock.patch('common.utils.flare.strftime', side_effect=mocked_strftime)
    @mock.patch('tempfile.gettempdir', side_effect=get_mocked_temp)
    @mock.patch('config.get_version', side_effect=get_mocked_version)
    @mock.patch('common.utils.flare.get_config', side_effect=get_mocked_config)
    def test_init(self, mock_config, mock_version, mock_tempdir, mock_strftime):
        f = Flare(case_id=1337)
        conf = mock_config()
        self.assertEqual(f._case_id, 1337)
        self.assertEqual(f._license_key, conf['license_key'])
        self.assertEqual(f._url, 'http://cidc.aiops.com/support/flare')
        self.assertEqual(f.tar_path, os.path.join(get_mocked_temp(), "CiAgent-1.tar.bz2"))

    @mock.patch('common.utils.flare.requests.post', return_value=FakeResponse())
    @mock.patch('config.get_version', side_effect=get_mocked_version)
    @mock.patch('common.utils.flare.strftime', side_effect=mocked_strftime)
    @mock.patch('tempfile.gettempdir', side_effect=get_mocked_temp)
    @mock.patch('common.utils.flare.get_config', side_effect=get_mocked_config)
    def test_upload_with_case(self, mock_config, mock_tempdir, mock_stfrtime, mock_version, mock_requests):
        f = Flare(case_id=1337)
        f._ask_for_email = lambda: 'test@example.com'

        assert not mock_requests.called
        f.upload()
        assert mock_requests.called
        args, kwargs = mock_requests.call_args_list[0]
        self.assertEqual(
            args,
            ('http://cidc.aiops.com/support/flare/1337?license_key=LICENSEKEY',)
        )
        self.assertEqual(
            kwargs['files']['flare_file'].name,
            os.path.join(get_mocked_temp(), "CiAgent-1.tar.bz2")
        )
        self.assertEqual(kwargs['data']['case_id'], 1337)
        self.assertEqual(kwargs['data']['email'], 'test@example.com')
        assert kwargs['data']['hostname']

    @mock.patch('common.utils.flare.requests.post', return_value=FakeResponse())
    @mock.patch('config.get_version', side_effect=get_mocked_version)
    @mock.patch('common.utils.flare.strftime', side_effect=mocked_strftime)
    @mock.patch('tempfile.gettempdir', side_effect=get_mocked_temp)
    @mock.patch('common.utils.flare.get_config', side_effect=get_mocked_config)
    def test_upload_no_case(self, mock_config, mock_tempdir, mock_stfrtime, mock_version, mock_requests):
        f = Flare()
        f._ask_for_email = lambda: 'test@example.com'

        assert not mock_requests.called
        f.upload()
        assert mock_requests.called
        args, kwargs = mock_requests.call_args_list[0]
        self.assertEqual(
            args,
            ('http://cidc.aiops.com/support/flare?license_key=LICENSEKEY',)
        )
        self.assertEqual(
            kwargs['files']['flare_file'].name,
            os.path.join(get_mocked_temp(), "CiAgent-1.tar.bz2")
        )
        self.assertEqual(kwargs['data']['case_id'], None)
        self.assertEqual(kwargs['data']['email'], 'test@example.com')
        assert kwargs['data']['hostname']

    @attr(requires='core_integration')
    @mock.patch('common.utils.flare.strftime', side_effect=mocked_strftime)
    @mock.patch('tempfile.gettempdir', side_effect=get_mocked_temp)
    @mock.patch('common.utils.flare.get_config', side_effect=get_mocked_config)
    def test_endpoint(self, mock_config, mock_temp, mock_stfrtime):
        f = Flare()
        f._ask_for_email = lambda: None
        # try:
        #     f.upload()
        #     raise Exception('Should fail before')
        # except Exception, e:
        #     self.assertEqual(str(e), "Your request is incorrect: Invalid inputs: 'API key unknown'")
