# stdlib
from datetime import datetime, timedelta
import unittest

# 3rd party
from nose.plugins.attrib import attr
import requests
import simplejson as json
from tornado.web import Application

# project
from forwarder.application import (
    #  APIServiceCheckTransaction,
    MAX_QUEUE_SIZE,
    THROTTLING_DELAY,
)
from forwarder.handlers import MetricTransaction
from forwarder.transaction import Transaction, TransactionManager


class memTransaction(Transaction):
    def __init__(self, size, manager):
        Transaction.__init__(self)
        self._trManager = manager
        self._size = size
        self._flush_count = 0

        self.is_flushable = False

    def flush(self):
        self._flush_count = self._flush_count + 1
        if self.is_flushable:
            self._trManager.tr_success(self)
        else:
            self._trManager.tr_error(self)

        self._trManager.flush_next()


@attr(requires='core_integration')
class TestTransaction(unittest.TestCase):

    def setUp(self):
        pass

    def testMemoryLimit(self):
        """Test memory limit as well as simple flush"""

        # No throttling, no delay for replay
        trManager = TransactionManager(timedelta(seconds=0), MAX_QUEUE_SIZE, timedelta(seconds=0))

        step = 10
        oneTrSize = (MAX_QUEUE_SIZE / step) - 1
        for i in xrange(step):
            tr = memTransaction(oneTrSize, trManager)
            trManager.append(tr)

        trManager.flush()

        # There should be exactly step transaction in the list, with
        # a flush count of 1
        self.assertEqual(len(trManager._transactions), step)
        for tr in trManager._transactions:
            self.assertEqual(tr._flush_count, 1)

        # Try to add one more
        tr = memTransaction(oneTrSize + 10, trManager)
        trManager.append(tr)

        # At this point, transaction one (the oldest) should have been removed from the list
        self.assertEqual(len(trManager._transactions), step)
        for tr in trManager._transactions:
            self.assertNotEqual(tr._id, 1)

        trManager.flush()
        self.assertEqual(len(trManager._transactions), step)
        # Check and allow transactions to be flushed
        for tr in trManager._transactions:
            tr.is_flushable = True
            # Last transaction has been flushed only once
            if tr._id == step + 1:
                self.assertEqual(tr._flush_count, 1)
            else:
                self.assertEqual(tr._flush_count, 2)

        trManager.flush()
        self.assertEqual(len(trManager._transactions), 0)

    def testThrottling(self):
        """Test throttling while flushing"""

        # No throttling, no delay for replay
        trManager = TransactionManager(timedelta(seconds=0), MAX_QUEUE_SIZE, THROTTLING_DELAY)
        trManager._flush_without_ioloop = True  # Use blocking API to emulate tornado ioloop

        # Add 3 transactions, make sure no memory limit is in the way
        oneTrSize = MAX_QUEUE_SIZE / 10
        for i in xrange(3):
            tr = memTransaction(oneTrSize, trManager)
            trManager.append(tr)

        # Try to flush them, time it
        before = datetime.now()
        trManager.flush()
        after = datetime.now()
        self.assertTrue((after - before) > 3 * THROTTLING_DELAY - timedelta(microseconds=100000),
                        "before = %s after = %s" % (before, after))

    def testCustomEndpoint(self):
        MetricTransaction._endpoints = []

        config = {
            "ci_url": "https://foo.bar.com",
            "license_key": "a" * 64
        }

        app = Application()
        app.skip_ssl_validation = False
        app._agentConfig = config
        app.use_simple_http_client = True

        trManager = TransactionManager(timedelta(seconds=0), MAX_QUEUE_SIZE, THROTTLING_DELAY)
        trManager._flush_without_ioloop = True  # Use blocking API to emulate tornado ioloop
        MetricTransaction._trManager = trManager
        MetricTransaction.set_application(app)
        MetricTransaction.set_endpoints()
        transaction = MetricTransaction(None, {}, "msgtype")
        endpoints = [transaction.get_url(e) for e in transaction._endpoints]
        expected = ["https://foo.bar.com/infrastructure/metrics?license_key=" + "a" * 64]
        self.assertEqual(endpoints, expected, (endpoints, expected))

    def testInvalidLicense(self):
        MetricTransaction._endpoints = []

        config = {
            "ci_url": "https://foo.bar.com",
            "license_key": ""
        }

        app = Application()
        app.skip_ssl_validation = False
        app._agentConfig = config
        app.use_simple_http_client = True

        trManager = TransactionManager(timedelta(seconds=0), MAX_QUEUE_SIZE, THROTTLING_DELAY)
        trManager._flush_without_ioloop = True  # Use blocking API to emulate tornado ioloop
        MetricTransaction._trManager = trManager
        MetricTransaction.set_application(app)
        MetricTransaction.set_endpoints()
        transaction = MetricTransaction(None, {}, "msgtype")
        endpoints = [transaction.get_url(e) for e in transaction._endpoints]
        expected = []
        self.assertEqual(endpoints, expected, (endpoints, expected))

    def testEndpoints(self):
        """
        Tests that the logic behind the agent version specific endpoints is ok.
        Also tests that these endpoints actually exist.
        """
        MetricTransaction._endpoints = []
        # license_key = "a" * 32
        license_key = "B1oCTldEV6b2bAoVCEwID0918f4FCQdXH1I9f9dBHApWFVQIcb79VVZPBw4bUAU="
        config = {
            "ci_url": "http://cidc.aiops.com",
            "license_key": license_key,
        }

        app = Application()
        app.skip_ssl_validation = False
        app._agentConfig = config
        app.use_simple_http_client = True

        trManager = TransactionManager(timedelta(seconds=0), MAX_QUEUE_SIZE, THROTTLING_DELAY)
        trManager._flush_without_ioloop = True  # Use blocking API to emulate tornado ioloop
        MetricTransaction._trManager = trManager
        MetricTransaction.set_application(app)
        MetricTransaction.set_endpoints()

        transaction = MetricTransaction(None, {}, "msgtype")
        endpoints = [transaction.get_url(e) for e in transaction._endpoints]
        expected = ['http://cidc.aiops.com/infrastructure/metrics?'
                    'license_key=B1oCTldEV6b2bAoVCEwID0918f4FCQdXH1I9f9dBHApWFVQIcb79VVZPBw4bUAU=']
        self.assertEqual(endpoints, expected, (endpoints, expected))

        for url in endpoints:
            r = requests.post(url, data=json.dumps({"foo": "bar"}),
                              headers={'Content-Type': "application/json"})
            r.raise_for_status()

        # FIXME: wait for the api supported
        # API Service Check Transaction
        # APIServiceCheckTransaction._trManager = trManager
        # APIServiceCheckTransaction.set_application(app)
        # APIServiceCheckTransaction.set_endpoints()

        # transaction = APIServiceCheckTransaction(None, {})
        # endpoints = [transaction.get_url(e) for e in transaction._endpoints]
        # expected = ['http://cidc.aiops.com/infrastructure/service_checks/?'
        #             'license_key=B1oCTldEV6b2bAoVCEwID0918f4FCQdXH1I9f9dBHApWFVQIcb79VVZPBw4bUAU=']
        # self.assertEqual(endpoints, expected, (endpoints, expected))

        # for url in endpoints:
        #     r = requests.post(url, data=json.dumps({'check': 'test', 'status': 0}),
        #                       headers={'Content-Type': "application/json"})
        #     r.raise_for_status()
