# -*- coding: utf-8 -*-
# pylint: disable=E0401
#
# Copyright © 2009-2010 CEA
# Pierre Raybaut
# Licensed under the terms of the CECILL License
# Modified for Cloudinsight

# stdlib
import logging
import os
import os.path as osp
import platform
# To manage the agent on OSX
from subprocess import (
    CalledProcessError,
    check_call,
    check_output,
)
import sys
import locale
import thread  # To manage the windows process asynchronously

# 3rd Party
# GUI Imports
from guidata.configtools import (
    add_image_path,
    get_family,
    get_icon,
    MONOSPACE,
)
from PyQt4.QtCore import (
    QPoint,
    QSize,
    Qt,
    QTimer,
    SIGNAL,
)
from guidata.qt.QtGui import (
    QApplication,
    QFont,
    QGroupBox,
    QHBoxLayout,
    QInputDialog,
    QLabel,
    QListWidget,
    QMenu,
    QMessageBox,
    QPushButton,
    QSplitter,
    QSystemTrayIcon,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)
from guidata.qthelpers import get_std_icon

# small hack to avoid having to patch the spyderlib library
# Needed because of py2exe bundling not being able to access
# the spyderlib image sources
import spyderlib.baseconfig
spyderlib.baseconfig.IMG_PATH = [""]
from spyderlib.widgets.sourcecode.codeeditor import CodeEditor

# 3rd Party others
import tornado.template as template
import yaml

# project
from config import (
    get_confd_path,
    get_config,
    get_config_path,
    get_logging_config,
    get_version
)
from common.check_status import (
    CollectorStatus,
    StatsdStatus,
    ForwarderStatus,
    logger_info,
)
from common.util import yLoader
from common.utils.flare import Flare
from common.utils.platform import Platform

# Constants describing the agent state
AGENT_RUNNING = 0
AGENT_START_PENDING = 1
AGENT_STOP_PENDING = 2
AGENT_STOPPED = 3
AGENT_UNKNOWN = 4

# Windows management
# Import Windows stuff only on Windows
if Platform.is_windows():
    import win32api
    import win32con
    import win32process
    import win32serviceutil
    import win32service

    # project
    from common.utils.pidfile import PidFile
    from common.utils.process import pid_exists

    WIN_STATUS_TO_AGENT = {
        win32service.SERVICE_RUNNING: AGENT_RUNNING,
        win32service.SERVICE_START_PENDING: AGENT_START_PENDING,
        win32service.SERVICE_STOP_PENDING: AGENT_STOP_PENDING,
        win32service.SERVICE_STOPPED: AGENT_STOPPED,
    }


log = logging.getLogger(__name__)

EXCLUDED_WINDOWS_CHECKS = [
    'docker_daemon',
    'kafka_consumer',
    'mcache',
    'mesos',
    'network',
    'zookeeper',
]

EXCLUDED_MAC_CHECKS = [
    'cacti',
    'iis',
    'sqlserver',
    'win32_event_log',
    'windows_service',
    'wmi_check',
]

MAIN_WINDOW_TITLE = "Cloudinsight Manager"

CI_SERVICE = "CiAgent"

HUMAN_SERVICE_STATUS = {
    AGENT_RUNNING: 'Agent is running',
    AGENT_START_PENDING: 'Agent is starting',
    AGENT_STOP_PENDING: 'Agent is stopping',
    AGENT_STOPPED: 'Agent is stopped',
    AGENT_UNKNOWN: "Cannot get Agent status",
}

REFRESH_PERIOD = 5000

OPEN_LOG = "Open log file"


def get_checks():
    checks = {}
    conf_d_directory = get_confd_path()

    for filename in sorted(os.listdir(conf_d_directory)):
        module_name, ext = osp.splitext(filename)
        if Platform.is_windows():
            excluded_checks = EXCLUDED_WINDOWS_CHECKS
        else:
            excluded_checks = EXCLUDED_MAC_CHECKS
        if filename.split('.')[0] in excluded_checks:
            continue
        if ext not in ('.yaml', '.example', '.disabled'):
            continue

        agent_check = AgentCheck(filename, ext, conf_d_directory)
        if (agent_check.enabled or agent_check.module_name not in checks or
           (not agent_check.is_example and not checks[agent_check.module_name].enabled)):
            checks[agent_check.module_name] = agent_check

    checks_list = checks.values()
    checks_list.sort(key=lambda c: c.module_name)

    return checks_list


class EditorFile(object):
    def __init__(self, file_path, description):
        self.file_path = file_path
        self.description = description

    def get_description(self):
        return self.description

    def save(self, content):
        try:
            f = open(self.file_path, 'w')
            f.write(content.encode('utf-8'))
            self.content = content
            info_popup("File saved.")
        except Exception, e:
            warning_popup("Unable to save file: \n %s" % str(e))
            raise


class CloudinsightConf(EditorFile):
    def __init__(self, config_path):
        EditorFile.__init__(self, config_path, "Agent settings file: CiAgent.conf")

    @property
    def license_key(self):
        config = get_config(parse_args=False, cfg_path=self.file_path)
        license_key = config.get('license_key', None)
        if not license_key or license_key == 'LICENSEKEYHERE':
            return None
        return license_key

    def check_license_key(self, editor):
        if self.license_key is None:
            license_key, ok = QInputDialog.getText(
                None, "Add your license key",
                "You must first set your license key in this file."
                " You can find it at http://ci.aiops.com/#/settings"
            )
            if ok and license_key:
                new_content = []
                for line in self.content.splitlines():
                    if "license_key:" in line:
                        new_content.append("license_key: %s" % str(license_key))
                    else:
                        new_content.append("%s" % line)
                new_content = "\n".join(new_content)
                self.save(new_content)
                editor.set_text(new_content)

                if agent_status() != AGENT_STOPPED:
                    agent_manager("restart")
                else:
                    agent_manager("start")
            elif not ok:
                warning_popup("The agent needs an license key to send metrics to Cloudinsight")
                if agent_status() != AGENT_STOPPED:
                    agent_manager("stop")
            else:
                self.check_license_key(editor)


class AgentCheck(EditorFile):
    def __init__(self, filename, ext, conf_d_directory):
        file_path = osp.join(conf_d_directory, filename)
        self.module_name = filename.split('.')[0]

        EditorFile.__init__(self, file_path, description=self.module_name.replace("_", " ").title())

        self.enabled = ext == '.yaml'
        self.is_example = ext == '.example'
        self.enabled_name = osp.join(conf_d_directory, "%s.yaml" % self.module_name)
        self.disabled_name = "%s.disabled" % self.enabled_name

    def enable(self):
        self.enabled = True
        os.rename(self.file_path, self.enabled_name)
        self.file_path = self.enabled_name

    def disable(self):
        self.enabled = False
        os.rename(self.file_path, self.disabled_name)
        self.file_path = self.disabled_name

    def save(self, content):
        check_yaml_syntax(content)
        EditorFile.save(self, content)


class PropertiesWidget(QWidget):
    def __init__(self, parent):
        QWidget.__init__(self, parent)
        font = QFont(get_family(MONOSPACE), 10, QFont.Normal)

        info_icon = QLabel()
        icon = get_std_icon('MessageBoxInformation').pixmap(24, 24)
        info_icon.setPixmap(icon)
        info_icon.setFixedWidth(32)
        info_icon.setAlignment(Qt.AlignTop)

        self.service_status_label = QLabel()
        self.service_status_label.setWordWrap(True)
        self.service_status_label.setAlignment(Qt.AlignTop)
        self.service_status_label.setFont(font)

        self.desc_label = QLabel()
        self.desc_label.setWordWrap(True)
        self.desc_label.setAlignment(Qt.AlignTop)
        self.desc_label.setFont(font)

        self.group_desc = QGroupBox("Description", self)
        layout = QHBoxLayout()
        layout.addWidget(info_icon)
        layout.addWidget(self.desc_label)
        layout.addStretch()
        layout.addWidget(self.service_status_label)

        self.group_desc.setLayout(layout)

        self.editor = CodeEditor(self)
        self.editor.setup_editor(linenumbers=True, font=font)
        self.editor.setReadOnly(False)
        self.group_code = QGroupBox("Source code", self)
        layout = QVBoxLayout()
        layout.addWidget(self.editor)
        self.group_code.setLayout(layout)

        self.enable_button = QPushButton(get_icon("apply.png"),
                                         "Enable", self)

        self.save_button = QPushButton(get_icon("filesave.png"),
                                       "Save", self)

        self.disable_button = QPushButton(get_icon("delete.png"),
                                          "Disable", self)

        self.refresh_button = QPushButton(get_icon("restart.png"),
                                          "Refresh", self)
        hlayout = QHBoxLayout()
        hlayout.addWidget(self.save_button)
        hlayout.addWidget(self.enable_button)
        hlayout.addWidget(self.disable_button)
        hlayout.addWidget(self.refresh_button)

        vlayout = QVBoxLayout()
        vlayout.addWidget(self.group_desc)
        vlayout.addWidget(self.group_code)
        self.html_window = HTMLWindow()
        vlayout.addWidget(self.html_window)

        vlayout.addLayout(hlayout)
        self.setLayout(vlayout)

        self.current_file = None

    def set_status(self):
        self.refresh_button.setEnabled(True)
        self.disable_button.setEnabled(False)
        self.enable_button.setEnabled(False)
        self.save_button.setEnabled(False)

    def set_item(self, check):
        self.refresh_button.setEnabled(False)
        self.save_button.setEnabled(True)
        self.current_file = check
        self.desc_label.setText(check.get_description())
        self.editor.set_text_from_file(check.file_path)
        check.content = self.editor.toPlainText().__str__()
        if check.enabled:
            self.disable_button.setEnabled(True)
            self.enable_button.setEnabled(False)
        else:
            self.disable_button.setEnabled(False)
            self.enable_button.setEnabled(True)

    def set_ci_conf(self, ci_conf):
        self.save_button.setEnabled(True)
        self.refresh_button.setEnabled(False)
        self.current_file = ci_conf
        self.desc_label.setText(ci_conf.get_description())
        self.editor.set_text_from_file(ci_conf.file_path)
        ci_conf.content = self.editor.toPlainText().__str__()
        self.disable_button.setEnabled(False)
        self.enable_button.setEnabled(False)
        ci_conf.check_license_key(self.editor)

    def set_log_file(self, log_file):
        self.save_button.setEnabled(False)
        self.refresh_button.setEnabled(True)
        self.disable_button.setEnabled(False)
        self.enable_button.setEnabled(False)
        scroll_position=0
        try:
            self.current_file = log_file
            self.desc_label.setText(log_file.get_description())
            try:
                with open(log_file.file_path, 'rb') as f:
                    text = f.read()
                    scroll_position=len(text.splitlines())
                self.editor.set_text(text.decode(locale.getpreferredencoding()))
            except Exception:
                self.editor.set_text_from_file(log_file.file_path)

            self.editor.go_to_line(scroll_position)
        except Exception as e:
            self.editor.set_text("Log file not found:%s"%e)


class HTMLWindow(QTextEdit):
    def __init__(self, parent=None):
        QTextEdit.__init__(self, parent)
        self.setReadOnly(True)
        self.setHtml(self.latest_status())

    def latest_status(self):
        try:
            loaded_template = template.Loader(".")
            statsd_status = StatsdStatus.load_latest_status()
            forwarder_status = ForwarderStatus.load_latest_status()
            collector_status = CollectorStatus.load_latest_status()
            generated_template = loaded_template.load("status.html").generate(
                port=22,
                platform=platform.platform(),
                agent_version=get_version(),
                python_version=platform.python_version(),
                logger_info=logger_info(),
                statsd=statsd_status.to_dict(),
                forwarder=forwarder_status.to_dict(),
                collector=collector_status.to_dict(),
            )
            return generated_template
        except Exception:
            return ("Unable to fetch latest status")


class MainWindow(QSplitter):
    def __init__(self, parent=None):
        prefix_conf = ''

        if Platform.is_windows():
            prefix_conf = 'windows_'

        log_conf = get_logging_config()

        QSplitter.__init__(self, parent)
        self.setWindowTitle(MAIN_WINDOW_TITLE)
        self.setWindowIcon(get_icon("agent.svg"))

        self.sysTray = SystemTray(self)

        self.connect(self.sysTray, SIGNAL("activated(QSystemTrayIcon::ActivationReason)"), self.__icon_activated)

        checks = get_checks()
        ci_conf = CloudinsightConf(get_config_path())
        self.create_logs_files_windows(log_conf, prefix_conf)

        listwidget = QListWidget(self)
        listwidget.addItems([osp.basename(check.module_name).replace("_", " ").title() for check in checks])

        self.properties = PropertiesWidget(self)

        self.setting_button = QPushButton(get_icon("info.png"),
                                          "Logs and Status", self)
        self.menu_button = QPushButton(get_icon("settings.png"),
                                       "Actions", self)
        self.settings = [
            ("Forwarder Logs", lambda: [self.properties.set_log_file(self.forwarder_log_file),
                self.show_html(self.properties.group_code, self.properties.html_window, False)]),
            ("Collector Logs", lambda: [self.properties.set_log_file(self.collector_log_file),
                self.show_html(self.properties.group_code, self.properties.html_window, False)]),
            ("Statsd Logs", lambda: [self.properties.set_log_file(self.statsd_log_file),
                self.show_html(self.properties.group_code, self.properties.html_window, False)]),
            ("JMX Fetch Logs", lambda: [self.properties.set_log_file(self.jmxfetch_log_file),
                self.show_html(self.properties.group_code, self.properties.html_window, False)]),
            ("Agent Status", lambda: [self.properties.html_window.setHtml(self.properties.html_window.latest_status()),
                self.show_html(self.properties.group_code, self.properties.html_window, True),
                self.properties.set_status()]),
        ]

        self.agent_settings = QPushButton(get_icon("edit.png"),
                                          "Settings", self)
        self.connect(self.agent_settings, SIGNAL("clicked()"),
                     lambda: [self.properties.set_ci_conf(ci_conf),
                     self.show_html(self.properties.group_code, self.properties.html_window, False)])

        self.setting_menu = SettingMenu(self.settings)
        self.connect(self.setting_button, SIGNAL("clicked()"),
                     lambda: self.setting_menu.popup(self.setting_button.mapToGlobal(QPoint(0, 0))))

        self.manager_menu = Menu(self)
        self.connect(self.menu_button, SIGNAL("clicked()"),
                     lambda: self.manager_menu.popup(self.menu_button.mapToGlobal(QPoint(0, 0))))

        holdingBox = QGroupBox("", self)
        Box = QVBoxLayout(self)
        Box.addWidget(self.agent_settings)
        Box.addWidget(self.setting_button)
        Box.addWidget(self.menu_button)
        Box.addWidget(listwidget)
        holdingBox.setLayout(Box)

        self.addWidget(holdingBox)
        self.addWidget(self.properties)

        self.connect(self.properties.enable_button, SIGNAL("clicked()"),
                     lambda: enable_check(self.properties))

        self.connect(self.properties.disable_button, SIGNAL("clicked()"),
                     lambda: disable_check(self.properties))

        self.connect(self.properties.save_button, SIGNAL("clicked()"),
                     lambda: save_file(self.properties))

        self.connect(self.properties.refresh_button, SIGNAL("clicked()"),
                     lambda: [self.properties.set_log_file(self.properties.current_file),
                     self.properties.html_window.setHtml(self.properties.html_window.latest_status())])

        self.connect(listwidget, SIGNAL('currentRowChanged(int)'),
                     lambda row: [self.properties.set_item(checks[row]),
                     self.show_html(self.properties.group_code, self.properties.html_window, False)])

        listwidget.setCurrentRow(0)

        self.setSizes([150, 1])
        self.setStretchFactor(1, 1)
        self.resize(QSize(950, 600))
        self.properties.set_ci_conf(ci_conf)

        self.do_refresh()

    def do_refresh(self):
        try:
            if self.isVisible():
                service_status = agent_status()
                self.properties.service_status_label.setText(HUMAN_SERVICE_STATUS[service_status])
        finally:
            QTimer.singleShot(REFRESH_PERIOD, self.do_refresh)

    def closeEvent(self, event):
        self.hide()
        self.sysTray.show()
        event.ignore()

    def __icon_activated(self, reason):
        if reason == QSystemTrayIcon.DoubleClick:
            self.show()

    def show_html(self, editor, html, state):
        if state is True:
            editor.setVisible(False)
            html.setVisible(True)
        else:
            editor.setVisible(True)
            html.setVisible(False)

    def create_logs_files_windows(self, config, prefix):
        self.forwarder_log_file = EditorFile(
            config.get('%sforwarder_log_file' % prefix),
            "Forwarder log file"
        )
        self.collector_log_file = EditorFile(
            config.get('%scollector_log_file' % prefix),
            "Collector log file"
        )
        self.statsd_log_file = EditorFile(
            config.get('%sstatsd_log_file' % prefix),
            "Statsd log file"
        )
        self.jmxfetch_log_file = EditorFile(
            config.get('jmxfetch_log_file'),
            "JMX log file"
        )

    def show(self):
        QSplitter.show(self)
        self.raise_()


class Menu(QMenu):
    ABOUT = "CiAgent v{0}"
    START = "Start"
    STOP = "Stop"
    RESTART = "Restart"
    FLARE = "Flare"
    MAC_LOGIN = '{0} at login'
    EXIT = "Exit"
    SYSTEM_EVENTS_CMD = 'tell application "System Events" to {0}'

    def __init__(self, parent=None):
        QMenu.__init__(self, parent)
        self.options = {}
        system_tray_menu = [
            (self.START, lambda: agent_manager("start")),
            (self.STOP, lambda: agent_manager("stop")),
            (self.RESTART, lambda: agent_manager("restart")),
        ]
        # First the version
        self.addAction(self.ABOUT.format(get_version())).setEnabled(False)
        self.addSeparator()

        for name, action in system_tray_menu:
            self.add_option(name, action)

        # enable or disable mac login
        if Platform.is_mac():
            self.add_option(self.MAC_LOGIN.format(self.enable_or_disable_mac()),
                            lambda: self.enable_or_disable_login())
        elif Platform.is_windows():
            self.add_option(self.FLARE, lambda: thread.start_new_thread(windows_flare, ()))

        # And finally the exit
        self.add_option(self.EXIT, lambda: sys.exit(0))

        self.connect(self, SIGNAL("aboutToShow()"), lambda: self.update_options())

    def add_option(self, name, action):
        menu_action = self.addAction(name)
        self.connect(menu_action, SIGNAL('triggered()'), action)
        self.options[name] = menu_action

    def enable_or_disable_mac(self):
        try:
            output = check_output(['osascript', '-e',
                                   self.SYSTEM_EVENTS_CMD.format('get the path of every login item whose name is "CiAgent"')])
            return 'Disable' if 'Cloudinsight' in output else 'Enable'
        except CalledProcessError, e:
            log.warning('Get login item failed with output:{0}'.format(e.output))
            return 'Disable'

    def enable_or_disable_login(self):
        previous = self.enable_or_disable_mac()
        if previous == 'Disable':
            command = 'delete every login item whose name is "CiAgent"'
        else:
            command = 'make login item at end with properties {path:"/Applications/CiAgent.app", name:"CiAgent", hidden:false}'
        try:
            check_call(['osascript', '-e', self.SYSTEM_EVENTS_CMD.format(command)])
            self.removeAction(self.options[self.MAC_LOGIN.format(previous)])
            self.removeAction(self.options[self.EXIT])
            self.add_option(self.MAC_LOGIN.format(self.enable_or_disable_mac()),
                            lambda: self.enable_or_disable_login())
            self.add_option(self.EXIT, lambda: sys.exit(0))
        except Exception, e:
            log.warning('Exception during Mac item login {0}: {1}'.format(previous, e))

    def update_options(self):
        status = agent_status()
        if status == AGENT_RUNNING:
            self.options[self.START].setEnabled(False)
            self.options[self.RESTART].setEnabled(True)
            self.options[self.STOP].setEnabled(True)
        elif status == AGENT_STOPPED:
            self.options[self.START].setEnabled(True)
            self.options[self.RESTART].setEnabled(False)
            self.options[self.STOP].setEnabled(False)
        elif status in [AGENT_START_PENDING, AGENT_STOP_PENDING]:
            self.options[self.START].setEnabled(False)
            self.options[self.RESTART].setEnabled(False)
            self.options[self.STOP].setEnabled(False)


class SettingMenu(QMenu):

    def __init__(self, settings, parent=None,):
        QMenu.__init__(self, parent)
        self.options = {}

        for name, action in settings:
            menu_action = self.addAction(name)
            self.connect(menu_action, SIGNAL('triggered()'), action)
            self.options[name] = menu_action


class SystemTray(QSystemTrayIcon):

    def __init__(self, parent=None):
        QSystemTrayIcon.__init__(self, parent)
        self.setIcon(get_icon("agent.png"))
        self.setVisible(True)
        self.options = {}

        menu = Menu(self.parent())
        self.setContextMenu(menu)


def disable_check(properties):
    check = properties.current_file
    new_content = properties.editor.toPlainText().__str__()

    if check.content != new_content:
        warning_popup("You must first save the file.")
        return

    properties.enable_button.setEnabled(True)
    properties.disable_button.setEnabled(False)
    check.disable()


def enable_check(properties):
    check = properties.current_file

    new_content = properties.editor.toPlainText().__str__()
    if check.content != new_content:
        warning_popup("You must first save the file")
        return

    properties.enable_button.setEnabled(False)
    properties.disable_button.setEnabled(True)
    check.enable()


def save_file(properties):
    current_file = properties.current_file
    new_content = properties.editor.toPlainText().__str__()
    current_file.save(new_content)


def check_yaml_syntax(content):
    try:
        yaml.load(content, Loader=yLoader)
    except Exception, e:
        warning_popup("Unable to parse yaml: \n %s" % str(e))
        raise


def service_manager(action):
    try:
        if action == 'stop':
            win32serviceutil.StopService(CI_SERVICE)
        elif action == 'start':
            win32serviceutil.StartService(CI_SERVICE)
        elif action == 'restart':
            win32serviceutil.RestartService(CI_SERVICE)
    except Exception, e:
        warning_popup("Couldn't %s service: \n %s" % (action, str(e)))


def service_manager_status():
    try:
        return WIN_STATUS_TO_AGENT[
            win32serviceutil.QueryServiceStatus(CI_SERVICE)[1]
        ]
    except Exception:
        return AGENT_UNKNOWN


def osx_manager(action):
    try:
        check_call(['CiAgent', action])
    except Exception, e:
        warning_popup("Couldn't execute CiAgent %s: \n %s" % (action, str(e)))


def osx_manager_status():
    try:
        check_output(['CiAgent', 'status'])
        return AGENT_RUNNING
    except CalledProcessError, e:
        if 'not running' in e.output:
            return AGENT_STOPPED
        elif 'STARTING' in e.output:
            return AGENT_START_PENDING
        elif 'STOPPED' in e.output:
            return AGENT_STOP_PENDING
        else:
            return AGENT_UNKNOWN


def agent_status():
    if Platform.is_windows():
        return service_manager_status()
    else:
        return osx_manager_status()


def agent_manager(action, async=True):
    if Platform.is_windows():
        manager = service_manager
    else:
        manager = osx_manager
    if not async:
        manager(action)
    else:
        thread.start_new_thread(manager, (action,))


def windows_flare():
    ok = question_popup("Your logs and configuration files are going to be collected."
                        " Are you sure you want to do this?")
    if not ok:
        info_popup("Flare cancelled")
        return
    f = Flare()
    f.collect()
    info_popup("Logs and configuration files have been collected."
               " Please send the following file to CiAgent:\n {0}".format(f.tar_path))


def warning_popup(message, parent=None):
    QMessageBox.warning(parent, 'Message', message, QMessageBox.Ok)


def info_popup(message, parent=None):
    QMessageBox.information(parent, 'Message', message, QMessageBox.Ok)


def question_popup(message, parent=None):
    result = QMessageBox.question(parent, 'Message', message, QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
    if result == QMessageBox.No:
        return False
    return True


def kill_old_process():
    """ Kills or brings to the foreground (if possible) any other instance of this program. It
    avoids multiple icons in the Tray on Windows. On OSX, we don't have to do anything: icons
    don't get duplicated. """
    # Is there another Agent Manager process running ?
    pidfile = PidFile('AgentManager').get_path()

    old_pid = None
    try:
        pf = file(pidfile, 'r')
        old_pid = int(pf.read().strip())
        pf.close()
    except (IOError, ValueError):
        pass

    if old_pid is not None and pid_exists(old_pid, 'CiAgentManager.exe'):
        try:
            handle = win32api.OpenProcess(win32con.PROCESS_ALL_ACCESS, False, old_pid)
            exe_path = win32process.GetModuleFileNameEx(handle, 0)

            # If (and only if) this process is indeed an instance of the GUI, let's kill it
            if 'CiAgentManager.exe' in exe_path:
                win32api.TerminateProcess(handle, -1)

            win32api.CloseHandle(handle)
        except win32api.error, e:
            msg = "Unable to open old pid: %s %s" % (old_pid, str(e))
            log.exception(msg)

    # If we reached that point it means the current process should be the only running
    # AgentManager.exe, let's save its pid
    pid = str(os.getpid())
    try:
        with open(pidfile, 'w+') as fp:
            fp.write(str(pid))
    except Exception, e:
        msg = "Unable to write pidfile: %s %s" % (pidfile, str(e))
        log.exception(msg)
        sys.stderr.write(msg + "\n")
        sys.exit(1)


if __name__ == '__main__':
    if Platform.is_windows():
        # Let's kill any other running instance of our GUI/SystemTray before starting a new one.
        kill_old_process()

    app = QApplication([])
    if Platform.is_mac():
        add_image_path(osp.join(os.getcwd(), 'images'))
        # add CiAgent in PATH
        os.environ['PATH'] = "{0}:{1}".format(
            '/opt/CiAgent/bin/',
            os.environ['PATH']
        )
        win = SystemTray() if len(sys.argv) < 2 else MainWindow()
    else:
        win = MainWindow()
    # Let's start the agent if he's not already started
    if agent_status() not in [AGENT_RUNNING, AGENT_START_PENDING]:
        agent_manager('start')
    win.show()
    app.exec_()
